/*
 *  Copyright (C) 2008-2011 The Android Open Source Project
 *  Copyright 2013 Sony Corporation
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
/*
 * ChangeLog:
 *   2012-2013 changed by Sony Corporation
 */
#define LOG_TAG "a2dp_audio_hw_icx"
/* #define LOG_NDEBUG 0 */
#include <cutils/log.h>


#include <errno.h>
#include <pthread.h>
#include <stdint.h>

#include <cutils/str_parms.h>

#include <hardware/hardware.h>
#include <system/audio.h>
#include <hardware/audio.h>

#include <hardware_legacy/power.h>

#include "liba2dp.h"
#include "android_audio_hw_icx_buffer.h"
#include "android_audio_hw_icx_stream.h"
#include "android_audio_hw_icx_mixer.h"
#include "android_audio_hw_icx_src.h"
#include "lib/timed_mute.h"

#include "android_audio_hw_icx_dump.h"

#include <sys/eventfd.h>
#include <sys/select.h>
#include <sys/resource.h>


/*----------------------------------------------------------------------------*/
#define ANDROID_PRIORITY_AUDIO (-19)
#define BT_HAL_THREAD_PRIORITY (30)

#define MAX_WRITE_RETRIES (5)

#define PARM_A2DP_SUSPENDED     "A2dpSuspended"
#define PARM_BLUETOOOTH_ENABLED "bluetooth_enabled"

#define A2DP_ADDR_FORMAT        "00:00:00:00:00:00"
#define A2DP_ADDR_FORMAT_LENGTH strlen(A2DP_ADDR_FORMAT)

#define THREAD_WAIT_TIME_NS 200000000 /* 200 msec */

#define DEVICE_MUTE_NAME         "device mute"
#define DEVICE_MUTE_FADE_TIME_MS (50)


/*--------------------------------------------------------------- singleton --*/
static audio_device* s_adev = NULL;

/*----------------------------------------------------------------------------*/
struct audio_device_ {
    struct audio_hw_device super;

    bool     is_bt_enabled;
    bool     is_bt_suspended;
    uint64_t last_write_time;

    /* thread */
    pthread_t       buf_thread;
    bool            buf_thread_exit;
    pthread_mutex_t device_lock;

    /* select */
    int    event_fd;
    fd_set fdset;

    /* stream and buffer */
    stream_out* output_std;
    ICXBuffer   buf_std;

    stream_out* output_icx;
    ICXBuffer   buf_icx;

    /* a2dp lib */
    a2dpData a2dp_data;
    char     a2dp_addr[20];
    bool     a2dp_active;

    /* mix */
    uint64_t* mix_buf;

    /* mute */
    struct timed_mute_handle* mute;
};

/*------------------------------------------------------------------- debug --*/
static void
log_icx_buffer(const audio_device* adev, const char* discriptor, const uint frame_num)
{
#if 0
    if (adev->output_std && adev->output_icx)
    {
        ALOGD("std [standby %s : read %5d : write %5d : read + write %5d] icx [standby %s : read %5d : write %5d : read + write %5d] %s %d",
              (adev->output_std->standby)?"true":"false",
              ICXB_get_readable_frame_num(&(adev->output_std->icxbuf)),
              ICXB_get_writable_frame_num(&(adev->output_std->icxbuf)),
              ICXB_get_readable_frame_num(&(adev->output_std->icxbuf)) + ICXB_get_writable_frame_num(&(adev->output_std->icxbuf)),
              (adev->output_icx->standby)?"true":"false",
              ICXB_get_readable_frame_num(&(adev->output_icx->icxbuf)),
              ICXB_get_writable_frame_num(&(adev->output_icx->icxbuf)),
              ICXB_get_readable_frame_num(&(adev->output_icx->icxbuf)) + ICXB_get_writable_frame_num(&(adev->output_icx->icxbuf)),
              discriptor,
              frame_num);
    }
#else
    (void)adev;
#endif
}

#if 0
#include "android_audio_hw_icx_debug.h"
static void log_open_parameters(audio_devices_t      devices,
                                audio_output_flags_t flags,
                                struct audio_config* config)
{
    ALOGV("audio_devices_t      => %s", debug_print_audio_devices(devices));
    ALOGV("audio_output_flags_t => %s", debug_print_audio_output_flags(flags));
    ALOGV("config => { sample_rate: %d, channel_mask: %s, format: %s }",
          config->sample_rate,
          debug_print_channel_mask(config->channel_mask),
          debug_print_audio_format(config->format));
}
#else
static void log_open_parameters(audio_devices_t      devices,
                                audio_output_flags_t flags,
                                struct audio_config* config)
{
    (void)devices;
    (void)flags;
    (void)config;
}
#endif


/*-------------------------------------------------------------------- a2dp --*/
char*
adev_get_a2dp_addr(audio_device* adev)
{
    return adev->a2dp_addr;
}

int
adev_set_a2dp_sink(audio_device* adev,
                   const char*   new_addr)
{
    if (new_addr != NULL)
    {
        if (strlen(new_addr) != A2DP_ADDR_FORMAT_LENGTH)
        {
            ALOGE("invalid addr [%s]", new_addr);
            return -1;
        }

        strlcpy(adev->a2dp_addr,
                new_addr,
                sizeof(adev->a2dp_addr));
    }

    a2dp_set_sink(adev->a2dp_data,
                  adev->a2dp_addr);

    ALOGD("a2dp_addr : %s", adev->a2dp_addr);

    return 0;
}

static int
initialize_a2dp(audio_device* adev,
                const char*   new_addr)
{
    if (adev->a2dp_active == true) { return 0; }

    ALOGD("%s() start", __FUNCTION__);

    int ret;
    ret = a2dp_init(44100, 2, &adev->a2dp_data);
    if (ret < 0)
    {
        ALOGE("a2dp_init failed ret: %d\n", ret);
        adev->a2dp_data = NULL;
        return ret;
    }

    ret = adev_set_a2dp_sink(adev, new_addr);
    if (ret != 0)
    {
        ALOGE("adev_set_a2dp_sink failed %d", ret);
        return ret;
    }

    adev->a2dp_active = true;
    ALOGD("%s() end", __FUNCTION__);
    return 0;
}

static void
stop_a2dp(audio_device* adev)
{
    if (adev->a2dp_active == false) { return; }

    ALOGD("%s() start", __FUNCTION__);
    a2dp_stop(adev->a2dp_data);
    adev->a2dp_active = false;
    ALOGD("%s() end", __FUNCTION__);

#ifdef ENABLE_DATA_DUMP
    close_dump_file(DTYPE_OUT);
#endif
}

static void
cleanup_a2dp(audio_device* adev)
{
    if (adev->a2dp_data)
    {
        HAL_ASSERT(adev->a2dp_active == false);

        a2dp_cleanup(adev->a2dp_data);
        adev->a2dp_data = NULL;
    }
}

/*------------------------------------------------------------ external LIB --*/
static int
initialize_MIX(audio_device* adev)
{
    int ret;
    ret = MW_initialize();
    if (ret != 0)
    {
        ALOGE("MW_initialize() failed %d", ret);
        return ret;
    }

    /* initialize buf for mix */
    adev->mix_buf = (uint64_t*)calloc(SIZE_OF_32BIT_PCM_FRAME,
                                      ICX_BUFFER_PERIOD_FRAME_NUM);
    if (adev->mix_buf == NULL)
    {
        ALOGE("failed to allocate mixer working buffer");
        return -1;
    }

    return 0;
}

static void
finalize_MIX(audio_device* adev)
{
    free(adev->mix_buf);
    adev->mix_buf = NULL;

    MW_finalize();
}

/*----------------------------------------------------------------------------*/
static int
initialize_MUTE(audio_device* adev)
{
    adev->mute = timed_mute_open(DEVICE_MUTE_NAME,
                                 44100,
                                 DEVICE_MUTE_FADE_TIME_MS);
    if (adev->mute == NULL)
    {
        ALOGE("timed_mute_open failed");
        timed_mute_close(adev->mute);
        return -1;
    }

    return 0;
}

static void
finalize_MUTE(audio_device* adev)
{
    timed_mute_close(adev->mute);
    adev->mute = NULL;
}

static int
exec_MUTE(audio_device* adev,
          uint64_t*     buffer,
          uint32_t      frame_num)
{
    ALOGV("%s() frame_num %d", __FUNCTION__, frame_num);

    int ret = timed_mute_attenuate(adev->mute,
                                   (int32_t*)buffer,
                                   (int)(frame_num * CH_NUM_OF_A_FRAME));
    HAL_ASSERT(ret == 0);       /* always success */

    return 0;
}

static void
set_fade_MUTE(audio_device* adev,
              uint32_t      mute_time_ms)
{
    if (adev->mute == NULL) { return; }

    timed_mute_switch(adev->mute, mute_time_ms);
}

static void
flush_MUTE(audio_device* adev)
{
    if (adev->mute == NULL) { return; }

    ALOGV("%s()", __FUNCTION__);
    timed_mute_flush(adev->mute);
}


/*----------------------------------------------------------------------------*/
static stream_out**
get_stream_out_ptr(audio_device* adev,
                   stream_type   stype)
{
    stream_out** out = NULL;
    switch(stype)
    {
    case STYPE_STD: out = &adev->output_std; break;
    case STYPE_ICX: out = &adev->output_icx; break;
    }
    return out;
}

/*------------------------------------------------------------------ select --*/
static int
initialize_select(audio_device* adev)
{
    adev->event_fd = eventfd(0, EFD_NONBLOCK);
    if (adev->event_fd == -1)
    {
        ALOGE("obtain event_fd failed !");
        return -1;
    }

    FD_ZERO(&(adev->fdset));
    FD_SET(adev->event_fd, &(adev->fdset));

    return 0;
}

static void
finalize_select(audio_device* adev)
{
    FD_CLR(adev->event_fd, &(adev->fdset));
}

static void
wakeup_select_wait(audio_device* adev)
{
    uint64_t value = 1;
    ssize_t  written_size = write(adev->event_fd, &value, sizeof(uint64_t));
}

static void
cleanup_select_wait(audio_device* adev)
{
    uint64_t value;
    ssize_t  readen = read(adev->event_fd, (void*)&value, sizeof(uint64_t));
}

static void
reset_select_wait(audio_device* adev)
{
    FD_ZERO(&(adev->fdset));
    FD_SET(adev->event_fd, &(adev->fdset));
}

static void
select_wait(audio_device*   adev,
            struct timeval* sleep_timeval)
{
    ALOGV("%s(%ld) start", __FUNCTION__, (sleep_timeval == NULL)?-1:sleep_timeval->tv_usec);

    int ret = select(adev->event_fd + 1, &(adev->fdset), NULL, NULL, sleep_timeval);
    ALOGE_IF(ret  < 0, "select wait error %d", ret);
    ALOGV_IF(ret == 0, "timeout");

    if (ret > 0)
    {
        cleanup_select_wait(adev);
    }
    reset_select_wait(adev);

    ALOGV("%s() end", __FUNCTION__);
}

static void
select_wait_until_wakeup(audio_device* adev)
{
    select_wait(adev, NULL);
}

static void
select_wait_with_timeout(audio_device* adev,
                         int32_t       sleep_time_us)
{
    struct timeval sleep_timeval;
    sleep_timeval.tv_sec  = 0;
    sleep_timeval.tv_usec = sleep_time_us;

    select_wait(adev, &sleep_timeval);
}


/*------------------------------------------------------------------ thread --*/
static int
create_thread(pthread_t* handle,
              int        priority,
              void*      (*entry)(void *),
              void*      arg)
{
    pthread_attr_t     attr;
    struct sched_param sp;

    int ret;

	ret = pthread_attr_init(&attr);
	if (ret < 0)
    {
        ALOGE("pthread_attr_init() error");
		return(ret);
	}

	ret = pthread_attr_setschedpolicy(&attr, SCHED_FIFO);
	if (ret < 0)
    {
        ALOGE("pthread_attr_setschedpolicy() error");
		pthread_attr_destroy(&attr);
		return(ret);
	}

	sp.sched_priority = priority;

	ret = pthread_attr_setschedparam(&attr, &sp);
	if (ret <0)
    {
        ALOGE("pthread_attr_setschedparam() error");
		pthread_attr_destroy(&attr);
		return(ret);
	}

    ret = pthread_create(handle,
                         (const pthread_attr_t *) &attr,
                         entry,
                         arg);
    pthread_attr_destroy(&attr);
    if (ret != 0)
    {
        ALOGE("starting thread failed: %d\n", ret);
        return -1;
    }

    pthread_setname_np((*handle), "a2dp_audio_hw_icx");

    return 0;
}

/*----------------------------------------------------------------------------*/
static void
adev_mutex_lock_streams_buf(audio_device* adev)
{
    ICXB_lock(&(adev->buf_icx));
    ICXB_lock(&(adev->buf_std));
}

static void
adev_mutex_unlock_streams_buf(audio_device* adev)
{
    ICXB_unlock(&(adev->buf_std));
    ICXB_unlock(&(adev->buf_icx));
}

/*----------------------------------------------------------------------------*/
static void
down_to_16(uint64_t* buffer,
           uint32_t  frame_num)
{
    uint i;
    uint64_t* src_buf = buffer;
    uint32_t* dst_buf = (uint32_t*)buffer;

    for (i = 0; i < frame_num; i++)
    {
        dst_buf[i] = (((src_buf[i] & 0xFFFF000000000000) >> 32) |
                      ((src_buf[i] & 0x00000000FFFF0000) >> 16));
    }
}

static uint
mix_streams(audio_device* adev,
            uint64_t*     out_buf,
            const uint    frame_num)
{
    ALOGV("%s() start : processed_frame_num = %d  --->", __FUNCTION__, processed_frame_num);

    HAL_ASSERT(frame_num <= ICX_BUFFER_PERIOD_FRAME_NUM);

    uint64_t icx_buf[ICX_BUFFER_PERIOD_FRAME_NUM];
    uint64_t std_buf[ICX_BUFFER_PERIOD_FRAME_NUM];
    memset(icx_buf, 0, sizeof(icx_buf));
    memset(std_buf, 0, sizeof(std_buf));

    int ret;
    uint32_t readable_icx = ICXB_get_readable_frame_num(&adev->buf_icx, false);
    if (readable_icx != 0)
    {
        HAL_ASSERT(frame_num <= readable_icx);
        ret = ICXB_read(&adev->buf_icx, frame_num, icx_buf, false);
        HAL_ASSERT(ret == ICXB_SUCCESS);
    }

    uint32_t readable_std = ICXB_get_readable_frame_num(&adev->buf_std, false);
    if (readable_std != 0)
    {
        HAL_ASSERT(frame_num <= readable_std);
        ret = ICXB_read(&adev->buf_std, frame_num, std_buf, false);
        HAL_ASSERT(ret == ICXB_SUCCESS);
    }


    uint mixed_frame_num = MW_mix(icx_buf, std_buf, out_buf, frame_num);
    HAL_ASSERT(mixed_frame_num <= frame_num);

    if (readable_icx != 0)
    {
        ICXB_increment_read_index(&adev->buf_icx, mixed_frame_num, false);
        if (sout_get_state(adev->output_icx) == SS_NOT_USED)
        {
            sout_flush_mute(adev->output_icx);
        }
    }
    if (readable_std != 0)
    {
        ICXB_increment_read_index(&adev->buf_std, mixed_frame_num, false);
        if (sout_get_state(adev->output_std) == SS_NOT_USED)
        {
            sout_flush_mute(adev->output_std);
        }
    }

    ALOGV("%s() end <---", __FUNCTION__);

    return mixed_frame_num;
}

static bool
adev_is_wait_capable(audio_device* adev)
{
    if ((system_time() - adev->last_write_time) < THREAD_WAIT_TIME_NS)
    {
        return true;            /* capable */
    }
    else
    {
        return false;
    }
}

/* usleep value */
static const int
sleep_table[SS_STATE_NUM][SS_STATE_NUM] =
{
    {    -1, 10000,     0,     0 },
    { 10000, 10000, 10000, 10000 },
    {     0, 10000,     0,     0 },
    {     0, 10000,     0,     0 }
};

/*
 * retval == 0 : no need to sleep. frame is available.
 * retval != 0 : sleep this time.
 */
static int
get_sleep_time(audio_device* adev)
{
    stream_state state_icx = sout_get_state(adev->output_icx);
    stream_state state_std = sout_get_state(adev->output_std);

    return sleep_table[state_icx][state_std];
}

static size_t
get_readable_frame_num(audio_device* adev)
{
    stream_state state_icx = sout_get_state(adev->output_icx);
    stream_state state_std = sout_get_state(adev->output_std);

    size_t readable_frame_num = 0;

    if (state_icx == SS_NOT_USED || state_std == SS_NOT_USED)
    {
        readable_frame_num = MAX(ICXB_get_readable_frame_num(&adev->buf_icx, false),
                                 ICXB_get_readable_frame_num(&adev->buf_std, false));
    }
    else
    {
        readable_frame_num = MIN(ICXB_get_readable_frame_num(&adev->buf_icx, false),
                                 ICXB_get_readable_frame_num(&adev->buf_std, false));
    }

    return readable_frame_num;
}

/* retval : abandoned frame num */
static uint
adjust(stream_out* out, ICXBuffer* icxbuf)
{
    if (out)
    {
        size_t readable_num = ICXB_get_readable_frame_num(icxbuf, false);

        if (sout_is_standby(out) &&
            (0 < readable_num && readable_num < PROCESSABLE_MIN_FRAME_NUM))
        {
            ICXB_increment_read_index(icxbuf, readable_num, false);
            return readable_num;
        }
    }
    return 0;
}

static void
adjust_unprocessable_frames(audio_device* adev)
{
    size_t adjusted_frame_num;
    adjusted_frame_num = adjust(adev->output_std, &adev->buf_std);
    if (adjusted_frame_num > 0)
    {
        ALOGV("std adjusted %d frames", adjusted_frame_num);
        sout_flush_mute(adev->output_std);
    }

    adjusted_frame_num = adjust(adev->output_icx, &adev->buf_icx);
    if (adjusted_frame_num > 0)
    {
        ALOGV("icx adjusted %d frames", adjusted_frame_num);
        sout_flush_mute(adev->output_icx);
    }
}

void
force_change_to_standby_if_readable_is_not_enough(stream_out* out,
                                                  ICXBuffer*  icxbuf)
{
    size_t readable_frame_num = ICXB_get_readable_frame_num(icxbuf, false);

    if (!sout_is_standby(out) && readable_frame_num < PROCESSABLE_MIN_FRAME_NUM)
    {
        sout_set_standby(out, true);
    }
}

static bool
adev_force_change_to_standby_if_wait_not_capable(audio_device* adev)
{
    stream_state state_icx = sout_get_state(adev->output_icx);
    stream_state state_std = sout_get_state(adev->output_std);
    if ((state_icx != SS_NOT_USED && state_std != SS_NOT_USED) &&
        !adev_is_wait_capable(adev))
    {
        ALOGE("Waiting limit overs. Force change to standby");
        ALOGE("---- icx %d (standby = %s) : std %d (standby = %s) %d ----",
              ICXB_get_readable_frame_num(&adev->buf_icx, false),
              sout_is_standby(adev->output_icx)?"true":"false",
              ICXB_get_readable_frame_num(&adev->buf_std, false),
              sout_is_standby(adev->output_std)?"true":"false",
              __LINE__);

        force_change_to_standby_if_readable_is_not_enough(adev->output_icx, &adev->buf_icx);
        force_change_to_standby_if_readable_is_not_enough(adev->output_std, &adev->buf_std);

        return true;
    }
    return false;
}

static void
device_sleep(audio_device* adev,
             int32_t       sleep_time_us)
{
    if (sleep_time_us == 0)
    {
        /* do nothing */
    }
    else if (sleep_time_us == -1)
    {
        select_wait_until_wakeup(adev);
    }
    else if (sleep_time_us > 0)
    {
        select_wait_with_timeout(adev, sleep_time_us);
    }
    else
    {
        ALOGE("invalid sleep_time %d", sleep_time_us);
    }
}

static uint32_t continuous_err = 0;
static uint32_t
write_to_a2dp(audio_device* adev,
              uint32_t*     src_buf,
              uint32_t      src_frame_num)
{
    uint32_t offset  = 0;
    int      retries = MAX_WRITE_RETRIES;

    while(1)
    {
        adev->a2dp_active = true;

        void*    buffer    = (void*)(src_buf + offset);
        uint32_t frame_num = (src_frame_num - offset);

        int processed_size = a2dp_write(adev->a2dp_data,
                                        buffer,
                                        frame_num * SIZE_OF_16BIT_PCM_FRAME);
        if (processed_size < 0)
        {
            ALOGI_IF(continuous_err == 0,
                     "a2dp_write failed (%d), skip %d frame and continue.",
                     processed_size, frame_num);
            continuous_err++;
            break;
        }
        else if (processed_size == 0)
        {
            ALOGI("a2dp_write could not process %d frame !!", frame_num);
            if (--retries == 0)
            {
                ALOGE("a2dp_write retry gave up !!!!!");
                break;
            }
            else
            {
                ALOGI("retry");
                usleep(50000);  /* 50 msec */
                continue;
            }
        }
        else                    /* success */
        {
            continuous_err = 0;

            adev->last_write_time = system_time();

            uint32_t processed_frame_num = processed_size / SIZE_OF_16BIT_PCM_FRAME;
            offset += processed_frame_num;

#ifdef ENABLE_DATA_DUMP
            open_dump_file(DTYPE_OUT, 44100, SIZE_OF_16BIT_PCM_FRAME);
            write_to_dump(DTYPE_OUT, buffer, processed_size);
#endif

            if (frame_num > processed_frame_num)
            {
                ALOGI("retry remaining frames");
                usleep(50000);  /* 50 msec */
                continue;
            }
            break;
        }
    }

    return offset;
}

static void*
_out_buf_thread_func(void *context)
{
    audio_device* adev = (audio_device*)context;
    size_t readable_frame_num;

    setpriority(PRIO_PROCESS, 0, ANDROID_PRIORITY_AUDIO);

    int32_t sleep_time = 0;

    while(!adev->buf_thread_exit)
    {
        device_sleep(adev, sleep_time);

        if (adev->buf_thread_exit)
        {
            ALOGV("%s() end",__FUNCTION__);
            return NULL;
        }

        pthread_mutex_lock(&adev->device_lock);

        adev_mutex_lock_streams_buf(adev);
        adjust_unprocessable_frames(adev);

        sleep_time = get_sleep_time(adev);
        if (sleep_time != 0)
        {
            cleanup_select_wait(adev);

            if (adev_force_change_to_standby_if_wait_not_capable(adev))
            {
                adev->last_write_time = system_time();
                sleep_time = 0;
            }
            adev_mutex_unlock_streams_buf(adev);
            goto wait;
        }
        else                    /* no need to sleep */
        {
            readable_frame_num = get_readable_frame_num(adev);
        }
        adev_mutex_unlock_streams_buf(adev);


        if (readable_frame_num >= PROCESSABLE_MIN_FRAME_NUM && !adev->buf_thread_exit)
        {
            adev_mutex_lock_streams_buf(adev);

            size_t processable_frame_num = UPTO(readable_frame_num,
                                                ICX_BUFFER_PERIOD_FRAME_NUM);

            uint mixed_frame_num = mix_streams(adev,
                                               adev->mix_buf,
                                               processable_frame_num);

            exec_MUTE(adev, adev->mix_buf, mixed_frame_num);
            down_to_16(adev->mix_buf, mixed_frame_num);

            pthread_cond_signal(&(adev->buf_icx.cond));
            pthread_cond_signal(&(adev->buf_std.cond));

            adev_mutex_unlock_streams_buf(adev);


            uint32_t processed = write_to_a2dp(adev,
                                               (uint32_t*)adev->mix_buf,
                                               mixed_frame_num);
            if (processed != mixed_frame_num)
            {
                pthread_mutex_unlock(&adev->device_lock);

                /* sleep for unprocessed frames. */
                usleep_by_frame_num(mixed_frame_num - processed, 44100);

                pthread_mutex_lock(&adev->device_lock);
            }


            adev_mutex_lock_streams_buf(adev);

            adjust_unprocessable_frames(adev);

            sleep_time = get_sleep_time(adev);
            if (sleep_time != 0)
            {
                cleanup_select_wait(adev);

                if (adev_force_change_to_standby_if_wait_not_capable(adev))
                {
                    adev->last_write_time = system_time();
                    sleep_time = 0;
                }
                adev_mutex_unlock_streams_buf(adev);
                goto wait;
            }
            else
            {
                readable_frame_num = get_readable_frame_num(adev);
            }
            adev_mutex_unlock_streams_buf(adev);
        }

wait:
        pthread_mutex_unlock(&adev->device_lock);

        /* TODO: should refactor.
           intend the case (std == AS_NOT_USED && icx == AS_NOT_USED). */
        if (sleep_time == -1)
        {
            flush_MUTE(adev);
            stop_a2dp(adev);
        }
    } /* while(!adev->buf_thread_exit) */

    ALOGV("%s() end",__FUNCTION__);
    return NULL;
}

/*----------------------------------------------------------------------------*/
static int
initialize_a2dp_and_start_thread(audio_device* adev)
{
    int ret;

    /* initialize a2dp */
    ret = initialize_a2dp(adev, "00:00:00:00:00:00");
    if (ret != 0)
    {
        ALOGE("initialize_a2dp failed: %d\n", ret);
        return ret;
    }

    /* start thread */
    ALOGD("start thread");
    adev->buf_thread_exit = false;
    ret = create_thread(&(adev->buf_thread),
                        BT_HAL_THREAD_PRIORITY,
                        _out_buf_thread_func,
                        (void*)adev);
    if (ret != 0)
    {
        ALOGE("create thraed failed");
        return ret;
    }

    return 0;
}

static bool
check_config_std(struct audio_config* config)
{
    /* accept only 44.1khz 16bit stereo */
    if ((config->format       != AUDIO_FORMAT_PCM_16_BIT)  ||
        (config->channel_mask != AUDIO_CHANNEL_OUT_STEREO) ||
        (config->sample_rate  != 44100))
    {
        ALOGE("invalid config");
        return false;
    }
    return true;
}

static int
adev_open_output_stream(struct audio_hw_device*   dev,
                        audio_io_handle_t         handle,
                        audio_devices_t           devices,
                        audio_output_flags_t      flags,
                        struct audio_config*      config,
                        struct audio_stream_out** astream_out_ptr)
{
    audio_device* adev = (audio_device *)dev;
    int ret;

    ALOGD("%s() ------------------------",__FUNCTION__);
    log_open_parameters(devices, flags, config);

    HAL_ASSERT(check_config_std(config));
    HAL_ASSERT(adev->output_std == NULL);
    HAL_ASSERT(adev->output_icx == NULL);

    acquire_wake_lock(PARTIAL_WAKE_LOCK, A2DP_WAKE_LOCK_NAME);


    pthread_mutex_lock(&adev->device_lock);

    ret = open_stream_std(&(adev->output_std),
                          &adev,
                          devices,
                          config,
                          &(adev->buf_std));
    if (ret != 0)
    {
        ALOGE("open_output_stream failed(%d).", ret);
        pthread_mutex_unlock(&adev->device_lock);
        return ret;
    }
    *astream_out_ptr = sout_get_audio_stream_out(adev->output_std);


    ret = initialize_a2dp_and_start_thread(adev);
    if (ret != 0)
    {
        ALOGE("initialize_a2dp_and_start_thread failed: %d\n", ret);
        close_stream_std(adev->output_std);
        adev->output_std = NULL;

        pthread_mutex_unlock(&adev->device_lock);
        return ret;
    }

    pthread_mutex_unlock(&adev->device_lock);

    ALOGD("%s() end --------------------",__FUNCTION__);

    return 0;
}

static bool
check_config_icx(struct audio_config* config)
{
    /* accept only 32bit stereo 44.1 to 192 kHz*/
    if ((config->format       != AUDIO_FORMAT_PCM_32_BIT)  ||
        (config->channel_mask != AUDIO_CHANNEL_OUT_STEREO))
    {
        ALOGE("invalid config");
        return false;
    }

    switch (config->sample_rate)
    {
    case  44100:
    case  48000:
    case  88200:
    case  96000:
    case 176400:
    case 192000:
        break;
    default:
        ALOGE("invalid sample rate");
        return false;
        break;
    }
    return true;
}

static int
adev_open_output_stream_icx(struct audio_hw_device*   dev,
                            audio_io_handle_t         handle,
                            audio_devices_t           devices,
                            audio_output_flags_t      flags,
                            struct audio_config*      config,
                            struct audio_stream_out** astream_out_ptr)
{
    audio_device* adev = (audio_device *)dev;
    int ret;

    ALOGD("%s() --------------------",__FUNCTION__);
    log_open_parameters(devices, flags, config);

    HAL_ASSERT(check_config_icx(config));
    HAL_ASSERT(adev->output_std != NULL);
    HAL_ASSERT(adev->output_icx == NULL);


    pthread_mutex_lock(&adev->device_lock);

    ret = open_stream_icx(&(adev->output_icx),
                          &adev,
                          devices,
                          config,
                          &(adev->buf_icx));
    if (ret != 0)
    {
        ALOGE("open_output_stream_icx failed(%d).", ret);
        pthread_mutex_unlock(&adev->device_lock);
        return ret;
    }
    *astream_out_ptr = sout_get_audio_stream_out(adev->output_icx);


    pthread_mutex_unlock(&adev->device_lock);

    ALOGD("%s() end ----------------",__FUNCTION__);
    return ret;
}

static void
finalize_a2dp_and_thread(audio_device* adev)
{
    /* finalize thread */
    if (adev->buf_thread)
    {
        ALOGD("finalize thread start");

        adev_mutex_lock_streams_buf(adev);
        adev->buf_thread_exit = true;
        wakeup_select_wait(adev);
        adev_mutex_unlock_streams_buf(adev);

        pthread_join(adev->buf_thread, (void **) NULL);

        ALOGD("finalize thread end");
    }

    /* finalize a2dp */
    ALOGD("finalize a2dp start");
    stop_a2dp(adev);
    cleanup_a2dp(adev);
    ALOGD("finalize a2dp end");
}

static void
close_stream(audio_device* adev,
             stream_type   stype)
{
    stream_out** out_ptr = get_stream_out_ptr(adev, stype);
    stream_out*  out     = *out_ptr;

    pthread_mutex_lock(&adev->device_lock);
    {
        /* finalize stream */
        switch(stype)
        {
        case STYPE_STD: close_stream_std(out); break;
        case STYPE_ICX: close_stream_icx(out); break;
        }
        *out_ptr = NULL;
    }
    pthread_mutex_unlock(&adev->device_lock);
}

static void
adev_close_output_stream(struct audio_hw_device*  dev,
                         struct audio_stream_out* stream)
{
    audio_device* adev = (audio_device*)dev;
    stream_out* out = (stream_out*)stream;

    ALOGD("%s() +++++++++++++++++++++++",__FUNCTION__);
    HAL_ASSERT(sout_is_standby(out) == true);

    close_stream(adev, STYPE_STD);
    HAL_ASSERT(adev->output_std == NULL);

    finalize_a2dp_and_thread(adev);

    release_wake_lock(A2DP_WAKE_LOCK_NAME);

    ALOGD("%s() end +++++++++++++++++++",__FUNCTION__);
}

static void
adev_close_output_stream_icx(struct audio_hw_device *dev,
                             struct audio_stream_out *stream)
{
    audio_device* adev = (audio_device*)dev;
    stream_out* out = (stream_out*)stream;

    ALOGD("%s() +++++++++++++++++++",__FUNCTION__);
    HAL_ASSERT(sout_is_standby(out) == true);

    close_stream(adev, STYPE_ICX);
    HAL_ASSERT(adev->output_icx == NULL);

    ALOGD("%s() end +++++++++++++++",__FUNCTION__);
}

static int
adev_set_parameters(struct audio_hw_device* dev,
                    const char*             kvpairs)
{
    audio_device* adev = (audio_device*)dev;
    struct str_parms* parms = str_parms_create_str(kvpairs);

    char value[8];
    int  ret;

    ret = GET_VALUE_FROM_STR_PARMS(parms, PARM_BLUETOOOTH_ENABLED, value);
    if (ret >= 0)
    {
        ALOGD("%s() %s", __FUNCTION__, kvpairs);
        pthread_mutex_lock(&adev->device_lock);

        adev->is_bt_enabled = (strcmp(value, "true") == 0)? true : false;

        pthread_mutex_unlock(&adev->device_lock);
        ALOGD("%s() end", __FUNCTION__);
    }


    ret = GET_VALUE_FROM_STR_PARMS(parms, PARM_A2DP_SUSPENDED, value);
    if (ret >= 0)
    {
        ALOGD("%s() %s", __FUNCTION__, kvpairs);
        pthread_mutex_lock(&adev->device_lock);

        adev->is_bt_suspended = (strcmp(value, "true") == 0)? true : false;
        HAL_ASSERT(adev->is_bt_suspended != true); /* suspend == true may passed only when interrupted by things like Skype.
                                                      may not occur in icx case. */
        pthread_mutex_unlock(&adev->device_lock);
        ALOGD("%s() end", __FUNCTION__);
    }


    ret = GET_VALUE_FROM_STR_PARMS(parms, "timed mute", value);
    if (ret >= 0)
    {
        ALOGV("%s() %s", __FUNCTION__, kvpairs);
        adev_set_fade(atoi(value));
        ALOGV("%s() end", __FUNCTION__);
    }

    ret = GET_VALUE_FROM_STR_PARMS(parms, "std timed mute", value);
    if (ret >= 0)
    {
        ALOGV("%s() %s", __FUNCTION__, kvpairs);
        adev_set_stream_fade(STYPE_STD, atoi(value));
        ALOGV("%s() end", __FUNCTION__);
    }

    ret = GET_VALUE_FROM_STR_PARMS(parms, "icx timed mute", value);
    if (ret >= 0)
    {
        ALOGV("%s() %s", __FUNCTION__, kvpairs);
        adev_set_stream_fade(STYPE_ICX, atoi(value));
        ALOGV("%s() end", __FUNCTION__);
    }

    str_parms_destroy(parms);

    return ret;
}

static char*
adev_get_parameters(const struct audio_hw_device* dev,
                    const char*                   keys)
{
    audio_device* adev = (audio_device*)dev;

    struct str_parms* parms     = str_parms_create_str(keys);
    struct str_parms* out_parms = str_parms_create();

    char value[8];
    int ret;

    ret = GET_VALUE_FROM_STR_PARMS(parms, PARM_BLUETOOOTH_ENABLED, value);
    if (ret >= 0)
    {
        ALOGD("%s() %s", __FUNCTION__, keys);
        pthread_mutex_lock(&adev->device_lock);

        str_parms_add_str(out_parms,
                          PARM_BLUETOOOTH_ENABLED,
                          adev->is_bt_enabled ? "true" : "false");

        pthread_mutex_unlock(&adev->device_lock);
        ALOGD("%s() end", __FUNCTION__);
    }

    ret = GET_VALUE_FROM_STR_PARMS(parms, PARM_A2DP_SUSPENDED, value);
    if (ret >= 0)
    {
        ALOGD("%s() %s", __FUNCTION__, keys);
        pthread_mutex_lock(&adev->device_lock);

        str_parms_add_str(out_parms,
                          PARM_A2DP_SUSPENDED,
                          adev->is_bt_suspended ? "true" : "false");

        pthread_mutex_unlock(&adev->device_lock);
        ALOGD("%s() end", __FUNCTION__);
    }

    char* str = str_parms_to_str(out_parms);
    str_parms_destroy(out_parms);
    str_parms_destroy(parms);

    return str;
}

static int adev_init_check(const struct audio_hw_device *dev)
{
    return 0;
}

static int adev_set_voice_volume(struct audio_hw_device *dev, float volume)
{
    return -ENOSYS;
}

static int adev_set_master_volume(struct audio_hw_device *dev, float volume)
{
    return -ENOSYS;
}

static int adev_set_mode(struct audio_hw_device *dev, audio_mode_t mode)
{
    return 0;
}

static int adev_set_mic_mute(struct audio_hw_device *dev, bool state)
{
    return -ENOSYS;
}

static int adev_get_mic_mute(const struct audio_hw_device *dev, bool *state)
{
    return -ENOSYS;
}

static size_t adev_get_input_buffer_size(const struct audio_hw_device *dev,
                                         const struct audio_config *config)
{
    return 0;
}

static int adev_open_input_stream(struct audio_hw_device *dev,
                                  audio_io_handle_t handle,
                                  audio_devices_t devices,
                                  struct audio_config *config,
                                  struct audio_stream_in **stream_in)
{
    return -ENOSYS;
}

static void adev_close_input_stream(struct audio_hw_device *dev,
                                    struct audio_stream_in *in)
{
    return;
}

static int adev_dump(const audio_hw_device_t *device, int fd)
{
    return 0;
}

static int
adev_close(hw_device_t* device)
{
    audio_device* adev = (audio_device*)device;

    ALOGD("%s()",__FUNCTION__);

    /* finalize streams */
    if (adev->output_icx)
    {
        adev_close_output_stream_icx(&(adev->super),
                                     sout_get_audio_stream_out(adev->output_icx));
    }
    if (adev->output_std)
    {
        adev_close_output_stream(&(adev->super),
                                 sout_get_audio_stream_out(adev->output_std));
    }

    /* finalize audio_device */
    SW_close();                 /* close handle for SRC */
    finalize_MUTE(adev);
    finalize_MIX(adev);
    ICXBuffer_finalize(&adev->buf_icx);
    ICXBuffer_finalize(&adev->buf_std);
    finalize_select(adev);
    pthread_mutex_destroy(&(adev->device_lock));

    free(adev);
    s_adev = NULL;

    ALOGD("%s() end",__FUNCTION__);

    return 0;
}

static uint32_t adev_get_supported_devices(const struct audio_hw_device *dev)
{
    return AUDIO_DEVICE_OUT_ALL_A2DP;
}

static void
initialize_hw_device_t(struct hw_device_t* hw_device,
                       const hw_module_t* module)
{
    hw_device->tag     = HARDWARE_DEVICE_TAG;
    hw_device->version = AUDIO_DEVICE_API_VERSION_1_0;
    hw_device->module  = (struct hw_module_t *) module;
    hw_device->close   = adev_close;
}

static void
initialize_audio_hw_device(struct audio_hw_device* ah_device,
                           const hw_module_t*      module)
{
    initialize_hw_device_t(&(ah_device->common), module);

    ah_device->get_supported_devices   = adev_get_supported_devices;
    ah_device->init_check              = adev_init_check;
    ah_device->set_voice_volume        = adev_set_voice_volume;
    ah_device->set_master_volume       = adev_set_master_volume;
    ah_device->set_mode                = adev_set_mode;
    ah_device->set_mic_mute            = adev_set_mic_mute;
    ah_device->get_mic_mute            = adev_get_mic_mute;
    ah_device->set_parameters          = adev_set_parameters;
    ah_device->get_parameters          = adev_get_parameters;
    ah_device->get_input_buffer_size   = adev_get_input_buffer_size;
    ah_device->open_output_stream      = adev_open_output_stream;
    ah_device->close_output_stream     = adev_close_output_stream;
    ah_device->open_input_stream       = adev_open_input_stream;
    ah_device->close_input_stream      = adev_close_input_stream;
    ah_device->dump                    = adev_dump;
    ah_device->open_output_stream_icx  = adev_open_output_stream_icx;
    ah_device->close_output_stream_icx = adev_close_output_stream_icx;
    /* ah_device->add_audio_effect               = dummy; */
    /* ah_device->remove_audio_effect            = dummy; */
}

static int
initialize_audio_device(audio_device*      adev,
                        const hw_module_t* module)
{
    int ret;

    initialize_audio_hw_device(&(adev->super), module);

    /* initialize audio_device */
    adev->is_bt_enabled   = true;
    adev->is_bt_suspended = false;
    adev->last_write_time = 0;

    /* thread */
    adev->buf_thread_exit = false;
    pthread_mutex_init(&adev->device_lock, NULL);

    /* select */
    ret = initialize_select(adev);
    if (ret != 0)
    {
        ALOGE("initialize_select failed !");
        return ret;
    }

    /* stream and buffer */
    adev->output_std = NULL;
    ret = ICXBuffer_initialize(&adev->buf_std,
                               ICX_BUFFER_PERIOD_FRAME_NUM);
    if (ret != 0)
    {
        ALOGE("ICXBuffer_initialize() failed");
        return ret;
    }

    adev->output_icx = NULL;
    ret = ICXBuffer_initialize(&adev->buf_icx,
                              ICX_BUFFER_PERIOD_FRAME_NUM);
    if (ret != 0)
    {
        ALOGE("ICXBuffer_initialize() failed");
        return ret;
    }

    /* a2dp lib */
    adev->a2dp_data   = NULL;
    memset(adev->a2dp_addr, 0, sizeof(adev->a2dp_addr));
    adev->a2dp_active = false;

    /* mix */
    ret = initialize_MIX(adev);
    if (ret != 0)
    {
        ALOGE("initialize_Mixer() failed !");
        return -1;
    }

    /* mute */
    ret = initialize_MUTE(adev);
    if (ret != 0)
    {
        ALOGE("initialize_MUTE failed");
        return -1;
    }

    /* open handle for SRC */
    ret = SW_open();
    if (ret != 0)
    {
        ALOGE("SW_open failed");
        return -1;
    }

    return 0;
}

static int
adev_open(const hw_module_t* module,
          const char*        name,
          hw_device_t**      device)
{
    ALOGD("%s(name = \"%s\")",__FUNCTION__, name);

    if (strcmp(name, AUDIO_HARDWARE_INTERFACE) != 0)
    {
        return -EINVAL;
    }

    if (s_adev == NULL)
    {
        s_adev = calloc(1, sizeof(audio_device));
        if (!s_adev)
        {
            return -ENOMEM;
        }

        int ret = initialize_audio_device(s_adev, module);
        if (ret != 0)
        {
            free(s_adev);
            return ret;
        }
    }

    *device = &s_adev->super.common;

    return 0;
}

static struct hw_module_methods_t hal_module_methods = {
    .open = adev_open,
};

struct audio_module HAL_MODULE_INFO_SYM = {
    .common = {
        .tag = HARDWARE_MODULE_TAG,
#ifdef AUDIO_MODULE_API_VERSION_0_1
        .module_api_version = AUDIO_MODULE_API_VERSION_0_1,
        .hal_api_version    = HARDWARE_HAL_API_VERSION,
#else
        .version_major = 1,
        .version_minor = 0,
#endif
        .id      = AUDIO_HARDWARE_MODULE_ID,
        .name    = "A2DP Audio HW HAL",
        .author  = "The Android Open Source Project",
        .methods = &hal_module_methods,
    },
};


/*------------------------------------------------------------------ public --*/
uint64_t adev_get_last_write_time(audio_device* adev)
{
    return adev->last_write_time;
}
void adev_set_last_write_time(audio_device* adev, uint64_t time)
{
    adev->last_write_time = time;
}
void adev_wakeup()
{
    wakeup_select_wait(s_adev);
}
void adev_set_fade(uint32_t mute_time_ms)
{
    if (s_adev == NULL) { return; }

    ALOGV("%s() mute_time_ms: %d", __FUNCTION__, mute_time_ms);
    set_fade_MUTE(s_adev, mute_time_ms);
}
void adev_set_stream_fade(stream_type stype, uint32_t mute_time_ms)
{
    if (s_adev == NULL) { return; }

    stream_out* out = *(get_stream_out_ptr(s_adev, stype));
    sout_set_fade(out, mute_time_ms);
}
