/*
 *  Copyright (C) 2008-2011 The Android Open Source Project
 *  Copyright 2013 Sony Corporation
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
/*
 * ChangeLog:
 *   2012-2013 changed by Sony Corporation
 */
#include "android_audio_hw_icx_buffer.h"

#define LOG_TAG "a2dp_audio_hw_buffer"
/* #define LOG_NDEBUG 0 */
#include <cutils/log.h>

#include <errno.h>

#include "android_audio_hw_icx.h"


/*----------------------------------------------------------------------------*/
ICXB_status
ICXBuffer_initialize(ICXBuffer* self,
                     size_t     frame_num_in_period_buffer)
{
    pthread_mutex_init(&(self->lock), NULL);
    pthread_cond_init(&(self->cond),  NULL);

    self->size_of_a_frame            = SIZE_OF_32BIT_PCM_FRAME;
    self->frame_num_in_period_buffer = frame_num_in_period_buffer;
    self->frame_num_in_buf           = frame_num_in_period_buffer * NUM_OF_PERIOD_BUFFER;
    self->buf                        = (uint64_t*)malloc(self->frame_num_in_buf * self->size_of_a_frame);

    ICXBuffer_reset(self);

    return (self->buf == NULL)?(ICXB_ERROR):(ICXB_SUCCESS);
}

void
ICXBuffer_finalize(ICXBuffer* self)
{
    pthread_cond_destroy(&(self->cond));
    pthread_mutex_destroy(&(self->lock));

    if (self->buf)
    {
        free(self->buf);
    }
}

void
ICXBuffer_reset(ICXBuffer* self)
{
    self->buf_read_index    = 0;
    self->buf_write_index   = 0;
    self->written_frame_num = 0;

    self->is_read_empty  = true;
    self->is_write_empty = false;
}

size_t
ICXBuffer_get_period_buffer_size(const ICXBuffer* self)
{
    return self->frame_num_in_period_buffer * self->size_of_a_frame;
}

size_t
ICXBuffer_get_period_buffer_frame_num(const ICXBuffer* self)
{
    return self->frame_num_in_period_buffer;
}

size_t
ICXBuffer_get_buffer_frame_num(const ICXBuffer* self)
{
    return ICXBuffer_get_period_buffer_frame_num(self) * NUM_OF_PERIOD_BUFFER;
}

uint64_t*
ICXBuffer_get_write_addr(ICXBuffer* self)
{
    return (self->buf + self->buf_write_index);
}

uint64_t*
ICXBuffer_get_read_addr(ICXBuffer* self)
{
    return (self->buf + self->buf_read_index);
}

/*----------------------------------------------------------------------------*/
uint32_t
ICXB_get_writable_frame_num(ICXBuffer* self, bool lock)
{
    uint32_t frame_num;

    if (lock) { pthread_mutex_lock(&self->lock); }

    if (self->buf_read_index == self->buf_write_index)
    {
        if (self->is_write_empty)
        {
            /* assert(self->is_read_empty == false); */
            frame_num = 0;
        }
        else
        {
            frame_num = self->frame_num_in_buf;
        }
    }
    else
    {
        if (self->buf_read_index < self->buf_write_index)
        {
            frame_num = (self->frame_num_in_buf - self->buf_write_index) + (self->buf_read_index);
        }
        else
        {
            frame_num = (self->buf_read_index - self->buf_write_index);
        }
    }

    if (lock) { pthread_mutex_unlock(&self->lock); }

    return frame_num;
}

uint32_t
ICXB_get_readable_frame_num(ICXBuffer* self, bool lock)
{
    uint32_t readable_frame_num;

    if (lock) { pthread_mutex_lock(&self->lock); }

    if (self->buf_read_index == self->buf_write_index)
    {
        if (self->is_read_empty)
        {
            /* assert(self->is_write_empty == false); */
            readable_frame_num = 0;
        }
        else
        {
            readable_frame_num = self->frame_num_in_buf;
        }
    }
    else
    {
        if (self->buf_read_index < self->buf_write_index)
        {
            readable_frame_num = (self->buf_write_index - self->buf_read_index);
        }
        else
        {
            readable_frame_num = (self->frame_num_in_buf - self->buf_read_index) + (self->buf_write_index);
        }
    }

    if (lock) { pthread_mutex_unlock(&self->lock); }

    return readable_frame_num;
}

static void
buf_write(ICXBuffer* self,
          uint64_t*  buffer,
          size_t     frame_num)
{
    memcpy(ICXBuffer_get_write_addr(self),
           buffer,
           frame_num * SIZE_OF_32BIT_PCM_FRAME);
}

static void
buf_read(ICXBuffer* self,
         uint64_t*  buffer,
         size_t     frame_num)
{
    memcpy(buffer,
           ICXBuffer_get_read_addr(self),
           frame_num * SIZE_OF_32BIT_PCM_FRAME);
}

ICXB_status
ICXB_write(ICXBuffer* self,
           size_t     frame_num,
           uint64_t*  buffer,
           bool       lock)
{
    if (lock) { pthread_mutex_lock(&self->lock); }

    uint32_t writable_frame_num = ICXB_get_writable_frame_num(self, false);

    if (writable_frame_num < frame_num)
    {
        ALOGV("writable_frame_num %d, frame_num %d", writable_frame_num, frame_num);
        if (lock) { pthread_mutex_unlock(&self->lock); }
        return ICXB_ERROR;
    }


    /* if using ICXB_get_writable_frame_num(), need to update flags at first.
       ICXB_get_writable_frame_num() depends on buf_write_index. */
    if (writable_frame_num == frame_num)
    {
        self->is_write_empty = true;
    }
    self->is_read_empty = false;


    if (self->buf_write_index + frame_num < self->frame_num_in_buf)
    {
        buf_write(self, buffer, frame_num);
        self->buf_write_index += frame_num;
    }
    else                        /* wrap around */
    {
        size_t size1 = self->frame_num_in_buf - self->buf_write_index;
        size_t size2 = (self->buf_write_index + frame_num) - self->frame_num_in_buf;
        /* assert(size1 + size2 == frame_num); */

        buf_write(self, buffer, size1);
        self->buf_write_index = 0;
        buffer += size1;

        buf_write(self, buffer, size2);
        self->buf_write_index = size2;
    }

    if (lock) { pthread_mutex_unlock(&self->lock); }

    return ICXB_SUCCESS;
}

ICXB_status
ICXB_read(ICXBuffer* self,
          size_t     frame_num,
          uint64_t*  buffer,
          bool       lock)
{
    if (lock) { pthread_mutex_unlock(&self->lock); }

    uint32_t readable_frame_num = ICXB_get_readable_frame_num(self, false);

    if (readable_frame_num < frame_num)
    {
        ALOGE("readable_frame_num %d, frame_num %d", readable_frame_num, frame_num);
        if (lock) { pthread_mutex_unlock(&self->lock); }
        return ICXB_ERROR;
    }

    if (self->buf_read_index + frame_num < self->frame_num_in_buf)
    {
        buf_read(self, buffer, frame_num);
    }
    else
    {
        uint64_t* tmp_buffer = buffer;

        size_t size1 = self->frame_num_in_buf - self->buf_read_index;
        size_t size2 = (self->buf_read_index + frame_num) - self->frame_num_in_buf;
        /* assert(size1 + size2 == frame_num); */

        /* NOTE:
           Do not use buf_read() for case wrap around.
           buf_read() depends on buf_read_index and it cannot be updated in ICXB_read(). */
        memcpy(tmp_buffer,
               self->buf + self->buf_read_index,
               size1 * SIZE_OF_32BIT_PCM_FRAME);

        tmp_buffer += size1;

        memcpy(tmp_buffer,
               self->buf,
               size2 * SIZE_OF_32BIT_PCM_FRAME);
    }

    if (lock) { pthread_mutex_unlock(&self->lock); }

    return ICXB_SUCCESS;
}


ICXB_status
ICXB_increment_read_index(ICXBuffer* self,
                          size_t     frame_num,
                          bool       lock)
{
    if (lock) { pthread_mutex_lock(&self->lock); }

    uint32_t readable_frame_num = ICXB_get_readable_frame_num(self, false);

    if (readable_frame_num < frame_num)
    {
        ALOGE("readable_frame_num %d, frame_num %d", readable_frame_num, frame_num);
        if (lock) { pthread_mutex_unlock(&self->lock); }
        return ICXB_ERROR;
    }

    /* if using ICXB_get_readable_frame_num(), need to update flags at first.
       ICXB_get_readable_frame_num() depends on buf_read_index. */
    if (readable_frame_num == frame_num)
    {
        self->is_read_empty = true;
    }
    self->is_write_empty = false;


    if (self->buf_read_index + frame_num < self->frame_num_in_buf)
    {
        self->buf_read_index += frame_num;
    }
    else
    {
        self->buf_read_index = (self->buf_read_index + frame_num) - self->frame_num_in_buf;
    }

    if (lock) { pthread_mutex_unlock(&self->lock); }
    return ICXB_SUCCESS;
}

void
ICXB_lock(ICXBuffer* self)
{
    HAL_ASSERT(self != NULL);
    pthread_mutex_lock(&(self->lock));
}

void
ICXB_unlock(ICXBuffer* self)
{
    HAL_ASSERT(self != NULL);
    pthread_mutex_unlock(&(self->lock));
}

ICXB_status
ICXB_wait_writable(ICXBuffer* self,
                   uint32_t   timeout_ms,
                   bool       lock)
{
    int ret = 0;

    if (lock) { pthread_mutex_lock(&self->lock); }

    uint32_t writable_frame_num = ICXB_get_writable_frame_num(self, false);
    if (writable_frame_num == 0)
    {
        ALOGV("no vacancy for write buffer");
        adev_wakeup();

        ret = pthread_cond_timeout_np(&self->cond, &self->lock, timeout_ms);
        if (ret == ETIMEDOUT || ret == -ETIMEDOUT)
        {
            ALOGD("timeout");
        }
    }

    if (lock) { pthread_mutex_unlock(&self->lock); }

    return ret;
}

