/*
 *  Copyright 2013 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU Lesser General Public License as published by
 *  the Free Software Foundation;  version 2.1 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 *
 */
/*----------------------------------------------------------------------------*/
#include "android_audio_hw_icx_dump.h"

#include <stdbool.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include <cutils/properties.h>

#define LOG_TAG "a2dp_audio_hw_dump"
/* #define LOG_NDEBUG 0 */
#include <cutils/log.h>

/*----------------------------------------------------------------------------*/
static int fd_std = -1;
static int fd_icx = -1;
static int fd_out = -1;

/*----------------------------------------------------------------------------*/
int*
get_fd(dump_type dtype)
{
    int* fd = NULL;
    switch (dtype)
    {
    case DTYPE_STD: { fd = &fd_std; } break;
    case DTYPE_ICX: { fd = &fd_icx; } break;
    case DTYPE_OUT: { fd = &fd_out; } break;
    }
    ALOGE_IF((fd == NULL), "%s() invalid dump_type %d", __FUNCTION__, dtype);
    return fd;
}

const char*
get_dtype_name(dump_type dtype)
{
    switch (dtype)
    {
    case DTYPE_STD: { return "STD"; } break;
    case DTYPE_ICX: { return "ICX"; } break;
    case DTYPE_OUT: { return "OUT"; } break;
    }
    return "unknown";
}

bool
is_data_dump_enabled(dump_type dtype)
{
    char value[256];
    int  rv;

	rv = property_get(DATA_DUMP_PROP, value, NULL);
	if(rv <= 0)
    {
        ALOGV("failed property_get() %d", rv);
		return false;
    }

    switch (dtype)
    {
    case DTYPE_STD:
        return (strstr(value, KEY_STD) != NULL) ? true : false;
        break;
    case DTYPE_ICX:
        return (strstr(value, KEY_ICX) != NULL) ? true : false;
        break;
    case DTYPE_OUT:
        return ((strstr(value, KEY_OUT) != NULL) || (strstr(value, "all") != NULL)) ? true : false;
        break;
    }

	return false;
}

/*----------------------------------------------------------------------------*/
void
open_dump_file(dump_type dtype, uint32_t sampling_rate, uint32_t frame_byte_size)
{
    if (!is_data_dump_enabled(dtype)) { return; }

    int* fd = get_fd(dtype);
    if (fd == NULL || *fd != -1)
    {
        ALOGV("dump_file_fd already opened");
        return;
    }

	char time_str[256];
    time_t sec;
	time(&sec);
	strftime(time_str, sizeof(time_str), "%Y-%m-%d.%H%M%S", localtime(&sec));

	char sampling_rate_str[8];
    snprintf(sampling_rate_str, sizeof(sampling_rate_str), "%d", sampling_rate);

	char sample_bit_size_str[8];
    snprintf(sample_bit_size_str, sizeof(sample_bit_size_str), "%d", (frame_byte_size * 8 / 2));

    char path[256];
    snprintf(path, sizeof(path),
             DATA_DUMP_FILE_FORMAT,
             time_str, sampling_rate_str, get_dtype_name(dtype), sample_bit_size_str);

    *fd = open(path, (O_RDWR | O_CREAT | O_TRUNC), 0666);
    if (*fd < 0)
    {
        ALOGE("open() failed");
        *fd = -1;
        return;
    }
}

void
close_dump_file(dump_type dtype)
{
    int* fd = get_fd(dtype);
    if (fd == NULL || *fd == -1) { return; }

    close(*fd);
    *fd = -1;
}

void
write_to_dump(dump_type dtype, const void* buffer, uint size)
{
    int* fd = get_fd(dtype);
    if (fd == NULL || *fd == -1) { return; }

    int ret = write(*fd, buffer, size);
    if (ret < 0) { ALOGE("write error %s dtype %d", strerror(errno), dtype); }
}
