/*
 *  Copyright (C) 2008-2011 The Android Open Source Project
 *  Copyright 2013 Sony Corporation
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */
/*
 * ChangeLog:
 *   2012-2013 changed by Sony Corporation
 */
#include "android_audio_hw_icx_stream.h"

#define LOG_TAG "a2dp_audio_hw_icx_stream"
/* #define LOG_NDEBUG 0 */

#include <cutils/log.h>
#include <cutils/str_parms.h>

#include "android_audio_hw_icx_common.h"
#include "android_audio_hw_icx_src.h"
#include "lib/timed_mute.h"

#include "android_audio_hw_icx_dump.h"

/*--------------------------------------------------------------- constants --*/
/* max wait time for buffer vacancy */
#define BUF_WRITE_AVAILABILITY_TIMEOUT_MS (5000)

/* for backward compatibility with older audio framework */
#ifndef AUDIO_PARAMETER_A2DP_SINK_ADDRESS
#define AUDIO_PARAMETER_A2DP_SINK_ADDRESS "a2dp_sink_address"
#endif

#define STREAM_MUTE_FADE_TIME_MS (50)

/*------------------------------------------------------------------- class --*/
struct stream_out_ {
    struct audio_stream_out super;

    stream_type     stype;
    audio_device*   parent;
    audio_devices_t device;
    bool            is_standby;
    pthread_mutex_t stream_lock;

    /* input pcm info */
    uint32_t       sample_rate;
    uint32_t       channels;
    audio_format_t format;

    /* Buffer */
    ICXBuffer* icxbuf;

    /* SRC */
    uint64_t*  write_work_buffer; /* Buffer for 32bit 2ch PCM.
                                     To store SRC output or bit size converted pcm. */

    /* MUTE */
    struct timed_mute_handle* mute;
};

/*------------------------------------------------------------ external LIB --*/
static void
finalize_SRC(stream_out* out)
{
    if (out->write_work_buffer)
    {
        free(out->write_work_buffer);
    }
    SW_finalize();
}

static int
initialize_SRC(stream_out* out)
{
    int ret;
    ret = SW_initialize(out->sample_rate, 44100);
    if (ret != 0)
    {
        ALOGE("SW_initialize failed");
        finalize_SRC(out);
        return -1;
    }

    /* work buffer for SRC */
    out->write_work_buffer
        = (uint64_t*)calloc(SW_get_max_output_frame_num(),
                            SIZE_OF_32BIT_PCM_FRAME);
    if (out->write_work_buffer == NULL)
    {
        ALOGE("failed to allocate write_work_buffer");
        finalize_SRC(out);
        return -1;
    }

    return 0;
}


static int
initialize_MUTE(stream_out* out,
                char*       name)
{
    out->mute = timed_mute_open(name, 44100, STREAM_MUTE_FADE_TIME_MS);
    if (out->mute == NULL)
    {
        ALOGE("mute open failed");
        timed_mute_close(out->mute);
        return -1;
    }

    return 0;
}

static void
finalize_MUTE(stream_out* out)
{
    timed_mute_close(out->mute);
}

static int
exec_MUTE(stream_out* out,
          uint64_t*   buffer,
          uint32_t    frame_num)
{
    int ret = timed_mute_attenuate(out->mute,
                                   (int*)buffer,
                                   frame_num * CH_NUM_OF_A_FRAME);
    HAL_ASSERT(ret == 0);       /* always success */
    return 0;
}

static void
set_fade_MUTE(stream_out* out,
              uint32_t    mute_time_ms)
{
    if (out->mute == NULL) { return; }

    timed_mute_switch(out->mute, mute_time_ms);
}

static void
flush_MUTE(stream_out* out)
{
    if (out->mute == NULL) { return; }

    ALOGV("%s()", __FUNCTION__);
    timed_mute_flush(out->mute);
}


/*---------------------------------------------------- stream_out interface --*/
static uint32_t out_get_sample_rate(const struct audio_stream *stream)
{
    const stream_out* out = (const stream_out*)stream;

    return out->sample_rate;
}

static int out_set_sample_rate(struct audio_stream *stream, uint32_t rate)
{
    stream_out* out = (stream_out*)stream;

    return 0;
}

uint32_t
get_frame_byte_size(stream_out* out)
{
    if (out == NULL)
    {
        ALOGE("%s() self is NULL", __FUNCTION__);
        return 0;
    }
    return audio_stream_frame_size(&out->super.common);
}

/* Return frame num which HAL wants to be written. */
static size_t out_get_buffer_size(const struct audio_stream *stream)
{
    stream_out* out = (stream_out*)stream;
    size_t buffer_size
        = ICXBuffer_get_period_buffer_frame_num(out->icxbuf) * get_frame_byte_size(out);

    return buffer_size;
}

static uint32_t out_get_channels(const struct audio_stream *stream)
{
    const stream_out* out = (const stream_out*)stream;

    return out->channels;
}

static audio_format_t out_get_format(const struct audio_stream *stream)
{
    const stream_out* out = (const stream_out*)stream;

    return out->format;
}

static int out_set_format(struct audio_stream *stream, audio_format_t format)
{
    return 0;
}

static int out_dump(const struct audio_stream *stream, int fd)
{
    return 0;
}

/* Return total buffer latency of HAL. RingBuffer contains with 32bit 44.1kHz. */
static uint32_t out_get_latency(const struct audio_stream_out *stream)
{
    const stream_out* out = (const stream_out*)stream;
    uint32_t latency = frame_num_to_msec(ICXBuffer_get_buffer_frame_num(out->icxbuf),
                                         44100);
    return latency;
}

static int out_set_volume(struct audio_stream_out *stream, float left,
                          float right)
{
    return -ENOSYS;
}

static int out_get_render_position(const struct audio_stream_out *stream,
                                   uint32_t *dsp_frames)
{
    return -ENOSYS;
}

static int
out_standby(struct audio_stream *stream)
{
    stream_out* out = (stream_out*)stream;

    ALOGD("%s() [%d]",__FUNCTION__, out->stype);

    sout_set_standby(out, true);

    ALOGD("%s() end",__FUNCTION__);

    return 0;
}

static int
out_set_parameters(struct audio_stream* stream,
                   const char*          kvpairs)
{
    stream_out* out = (stream_out*)stream;

    struct str_parms* parms = str_parms_create_str(kvpairs);

    char value[32];

    int ret;
    ret = GET_VALUE_FROM_STR_PARMS(parms, AUDIO_PARAMETER_A2DP_SINK_ADDRESS, value);
    if (ret >= 0)
    {
        ALOGD("%s() %s", __FUNCTION__, kvpairs);
        pthread_mutex_lock(&out->stream_lock);

        ret = adev_set_a2dp_sink(out->parent, value);
        HAL_ASSERT(ret == 0);

        pthread_mutex_unlock(&out->stream_lock);
        ALOGD("%s() end",__FUNCTION__);
    }


    ret = GET_VALUE_FROM_STR_PARMS(parms, "timed mute", value);
    if (ret >= 0)
    {
        ALOGD("%s() %s", __FUNCTION__, kvpairs);
        adev_set_fade(atoi(value));
        ALOGD("%s() end",__FUNCTION__);
    }

    ret = GET_VALUE_FROM_STR_PARMS(parms, "std timed mute", value);
    if (ret >= 0)
    {
        ALOGD("%s() %s", __FUNCTION__, kvpairs);
        adev_set_stream_fade(STYPE_STD, atoi(value));
        ALOGD("%s() end",__FUNCTION__);
    }

    ret = GET_VALUE_FROM_STR_PARMS(parms, "icx timed mute", value);
    if (ret >= 0)
    {
        ALOGD("%s() %s", __FUNCTION__, kvpairs);
        adev_set_stream_fade(STYPE_ICX, atoi(value));
        ALOGD("%s() end",__FUNCTION__);
    }

    str_parms_destroy(parms);

    return ret;
}

static audio_devices_t out_get_device(const struct audio_stream *stream)
{
    const stream_out* out = (const stream_out*)stream;

    return out->device;
}

static int out_set_device(struct audio_stream *stream, audio_devices_t device)
{
    stream_out* out = (stream_out*)stream;

    if (!audio_is_a2dp_device(device))
    {
        return -EINVAL;
    }

    out->device = device;
    return 0;
}

static char*
out_get_parameters(const struct audio_stream* stream,
                   const char*                keys)
{
    stream_out* out = (stream_out*)stream;

    struct str_parms* parms     = str_parms_create_str(keys);
    struct str_parms* out_parms = str_parms_create();

    char value[20];
    int ret;

    ret = GET_VALUE_FROM_STR_PARMS(parms, AUDIO_PARAMETER_A2DP_SINK_ADDRESS, value);
    if (ret >= 0)
    {
        ALOGD("%s() %s", __FUNCTION__, keys);
        pthread_mutex_lock(&out->stream_lock);

        str_parms_add_str(out_parms,
                          AUDIO_PARAMETER_A2DP_SINK_ADDRESS,
                          adev_get_a2dp_addr(out->parent));

        pthread_mutex_unlock(&out->stream_lock);
        ALOGD("%s() end",__FUNCTION__);
    }

    char* str = str_parms_to_str(out_parms);
    str_parms_destroy(out_parms);
    str_parms_destroy(parms);

    return str;
}

static ssize_t
out_write_main(stream_out* out,
               size_t      buf_frames_num,
               uint64_t*   buf)
{
    /* ALOGV("%s(buf_frames_num %d)",__FUNCTION__, buf_frames_num); */

    ICXB_lock(out->icxbuf);
    {
        if (sout_is_standby(out))
        {
            ALOGD("first write [%d]", out->stype);
            sout_set_standby(out, false);
            ICXBuffer_reset(out->icxbuf);

            if (adev_get_last_write_time(out->parent) == 0)
            {
                adev_set_last_write_time(out->parent, system_time());
            }
        }
    }
    ICXB_unlock(out->icxbuf);


    exec_MUTE(out, buf, buf_frames_num);


    size_t written_frames_num = 0;
    while (written_frames_num < buf_frames_num)
    {
        ICXB_lock(out->icxbuf);

        ICXB_status ret = ICXB_wait_writable(out->icxbuf, BUF_WRITE_AVAILABILITY_TIMEOUT_MS, false);
        if (ret != ICXB_SUCCESS)
        {
            ALOGE("ICXB_wait_writable() can't wait");
            ICXB_unlock(out->icxbuf);
            return written_frames_num;
        }

        size_t writable_frames_num = ICXB_get_writable_frame_num(out->icxbuf, false);
        size_t rest_frames_num     = buf_frames_num - written_frames_num;

        size_t frames = UPTO(writable_frames_num, rest_frames_num);

        ret = ICXB_write(out->icxbuf, frames, buf + written_frames_num, false);
        HAL_ASSERT(ret == ICXB_SUCCESS);

        written_frames_num += frames;
        adev_wakeup();

        ICXB_unlock(out->icxbuf);
    }

    return 0;
}

static ssize_t
src_and_write_to_buffer(stream_out* stream,
                        const void* input_buf,
                        const uint  input_frame_num)
{
    ALOGV("%s() start input_frame_num => %d ----", __FUNCTION__, input_frame_num);
    HAL_ASSERT(input_frame_num >= SRC_PROCESSING_FRAME_NUM);
    HAL_ASSERT(input_frame_num % SRC_PROCESSING_FRAME_NUM == 0);

    int  ret;
    uint processed_frame_num = 0;
    int  src_output_frame_num;
    while (processed_frame_num < input_frame_num)
    {
        src_output_frame_num = 0;

        /* exec SRC */
        ret = SW_convert((uint64_t*)input_buf,
                         stream->write_work_buffer,
                         &src_output_frame_num);
        if (ret != 0)
        {
            ALOGE("SW_convert() failed !! %d", ret);
            return ret;
        }

        /* store in ICXBuffer */
        ret = out_write_main(stream,
                             src_output_frame_num,
                             stream->write_work_buffer);
        if (ret != 0)
        {
            ALOGV("result in out_write_main() %d", ret);
            return ret;
        }

        input_buf = ((uint64_t*)input_buf) + SRC_PROCESSING_FRAME_NUM;
        processed_frame_num += SRC_PROCESSING_FRAME_NUM;
    }

    /* ALOGV("%s() end <<<<", __FUNCTION__); */

    return 0;
}

/**
 * @param in_buf    [in]  16bit 2ch PCM buffer.
 * @param frame_num [in]  Number of frames contained in in_buf.
 * @param out_buf   [out] 32bit 2ch PCM buffer.
 */
static void
up_to_32bit(const uint32_t* in_buf,
            const size_t    frame_num,
            uint64_t*       out_buf)
{
    uint i;
    uint16_t sample16_left;
    uint16_t sample16_right;

    for (i = 0; i < frame_num; i++)
    {
        sample16_left  = (in_buf[i] >> 16);
        sample16_right = (in_buf[i] << 16) >> 16;

        out_buf[i] = sample16_left;                         /* 0x0000-0000-0000-LEFT */
        out_buf[i] = (out_buf[i] << 32) | (sample16_right); /* 0x0000-LEFT-0000-RIGH */
        out_buf[i] <<= 16;                                  /* 0xLEFT-0000-RIGH-0000 */
    }
}

static ssize_t
out_write(struct audio_stream_out* stream,
          const void*              buffer,
          size_t                   bytes)
{
    ALOGV("%s(bytes = %d) start >>>>>>>>>>>>>>>>>>>>",__FUNCTION__, bytes);

    stream_out* out = (stream_out*)stream;

    size_t frames_total = bytes / audio_stream_frame_size(&out->super.common);
    HAL_ASSERT(frames_total <= ICX_BUFFER_PERIOD_FRAME_NUM);

#ifdef ENABLE_DATA_DUMP
    open_dump_file(DTYPE_STD, out->sample_rate, SIZE_OF_16BIT_PCM_FRAME);
    write_to_dump(DTYPE_STD, buffer, bytes);
#endif

    up_to_32bit((uint32_t*)buffer,
                frames_total,
                out->write_work_buffer);

    int ret = out_write_main(out, frames_total, out->write_work_buffer);
    if (ret != 0)
    {
        ALOGV("%s(bytes = %d) end <<<<<<<<<<<<<<<<<<<<",__FUNCTION__, bytes);
        return ret;
    }

    ALOGV("%s(bytes = %d) end <<<<<<<<<<<<<<<<<<<<",__FUNCTION__, bytes);
    return bytes;
}

static ssize_t
out_write_icx(struct audio_stream_out* stream,
              const void*              buffer,
              size_t                   bytes)
{
    ALOGV("%s(bytes = %d) start >>>>>>>>>>>>>>>>>>>>",__FUNCTION__, bytes);

    stream_out* out = (stream_out*)stream;

    size_t frames_total = bytes / audio_stream_frame_size(&(out->super.common));

#ifdef ENABLE_DATA_DUMP
    open_dump_file(DTYPE_ICX, out->sample_rate, SIZE_OF_32BIT_PCM_FRAME);
    write_to_dump(DTYPE_ICX, buffer, bytes);
#endif

    int ret = src_and_write_to_buffer(out, buffer, frames_total);
    if (ret != 0)
    {
        ALOGV("%s(bytes = %d) end <<<<<<<<<<<<<<<<<<<<",__FUNCTION__, bytes);
        return ret;
    }

    ALOGV("%s(bytes = %d) end <<<<<<<<<<<<<<<<<<<<",__FUNCTION__, bytes);
    return bytes;
}

static int out_add_audio_effect(const struct audio_stream *stream, effect_handle_t effect)
{
    return 0;
}

static int out_remove_audio_effect(const struct audio_stream *stream, effect_handle_t effect)
{
    return 0;
}

/*-------------------------------------------------------------- initializer--*/
static void
initialize_audio_stream(struct audio_stream* as)
{
    as->get_sample_rate     = out_get_sample_rate;
    as->set_sample_rate     = out_set_sample_rate;
    as->get_buffer_size     = out_get_buffer_size;
    as->get_channels        = out_get_channels;
    as->get_format          = out_get_format;
    as->set_format          = out_set_format;
    as->standby             = out_standby;
    as->dump                = out_dump;
    as->set_parameters      = out_set_parameters;
    as->get_parameters      = out_get_parameters;
    as->set_device          = out_set_device;
    as->get_device          = out_get_device;
    as->add_audio_effect    = out_add_audio_effect;
    as->remove_audio_effect = out_remove_audio_effect;
}

typedef ssize_t (*write_func_ptr)(struct audio_stream_out *stream,
                                  const void* buffer,
                                  size_t bytes);

static void
initialize_audio_stream_out(struct audio_stream_out* as_out,
                            write_func_ptr           write_func)
{
    /* initialize audio_stream */
    initialize_audio_stream(&(as_out->common));

    /* initialize audio_stream_out */
    as_out->get_latency         = out_get_latency;
    as_out->set_volume          = out_set_volume;
    as_out->write               = write_func;
    as_out->get_render_position = out_get_render_position;
    /* as_out->get_next_write_timestamp = dummy; */
}

static int
initialize_stream_out(stream_out*          out,
                      stream_type          stype,
                      audio_device**       parent,
                      audio_devices_t      devices,
                      struct audio_config* config,
                      ICXBuffer*           buf,
                      write_func_ptr       write_func)
{
    /* initialize audio_stream_out */
    initialize_audio_stream_out(&out->super, write_func);

    /* initialize variables */
    out->stype       = stype;
    out->parent      = *parent;
    out->device      = devices;
    out->is_standby  = true;
    pthread_mutex_init(&out->stream_lock, NULL);
    out->sample_rate = config->sample_rate;
    out->channels    = config->channel_mask;
    out->format      = config->format;
    out->icxbuf      = buf;
    ICXBuffer_reset(out->icxbuf);

    /* mute */
    int ret = initialize_MUTE(out,
                              (stype == STYPE_STD) ? "std mute" : "icx mute");
    if (ret != 0)
    {
        ALOGE("initialize_MUTE failed");
        return -1;
    }

    return 0;
}

static void
finalize_stream_out(stream_out* out)
{
    finalize_MUTE(out);
}

static int
create_stream_out(stream_out**         stream_out_ptr,
                  stream_type          stype,
                  audio_device**       parent,
                  audio_devices_t      devices,
                  struct audio_config* config,
                  ICXBuffer*           buf,
                  write_func_ptr       write_func)
{
    /* alloc stream_out. */
    stream_out* out = calloc(1, sizeof(stream_out));
    if (!out)
    {
        return -ENOMEM;
    }

    /* initialize stream_out. */
    int ret = initialize_stream_out(out,
                                    stype,
                                    parent,
                                    devices,
                                    config,
                                    buf,
                                    write_func);
    if (ret != 0)
    {
        free(out);
        return ret;
    }

    *stream_out_ptr = out;

    config->format       = out->format;
    config->channel_mask = out->channels;
    config->sample_rate  = out->sample_rate;

    return 0;
}

/*----------------------------------------------------------------------------*/
int
open_stream_std(stream_out**         stream_out_ptr,
                audio_device**       parent,
                audio_devices_t      devices,
                struct audio_config* config,
                ICXBuffer*           buf)
{
    int ret;
    /* create stream_out into adev->output_std. */
    ret = create_stream_out(stream_out_ptr,
                            STYPE_STD,
                            parent,
                            devices,
                            config,
                            buf,
                            out_write);
    if (ret != 0)
    {
        ALOGE("create_output_stream failed(%d).", ret);
        return ret;
    }

    /* work buffer for 16bit to 32bit convert */
    (*stream_out_ptr)->write_work_buffer
        = (uint64_t*)calloc(ICX_BUFFER_PERIOD_FRAME_NUM,
                            SIZE_OF_32BIT_PCM_FRAME);
    if ((*stream_out_ptr)->write_work_buffer == NULL)
    {
        ALOGE("failed to allocate write_work_buffer");
        free((*stream_out_ptr));
        return -1;
    }

    return 0;
}

int
open_stream_icx(stream_out**         stream_out_ptr,
                audio_device**       parent,
                audio_devices_t      devices,
                struct audio_config* config,
                ICXBuffer*           buf)
{
    int ret;

    /* create stream_out into adev->output_icx. */
    ret = create_stream_out(stream_out_ptr,
                            STYPE_ICX,
                            parent,
                            devices,
                            config,
                            buf,
                            out_write_icx);
    if (ret != 0)
    {
        ALOGE("create_output_stream failed(%d).", ret);
        return ret;
    }

    /* SRC only used in icx path */
    ret = initialize_SRC((*stream_out_ptr));
    if (ret != 0)
    {
        ALOGE("failed to initialize SRConv.");
        free((*stream_out_ptr));
        return ret;
    }

    return 0;
}

void close_stream_std(stream_out* out)
{
    ALOGD("%s() start", __FUNCTION__);

#ifdef ENABLE_DATA_DUMP
    close_dump_file(DTYPE_STD);
#endif

    finalize_stream_out(out);
    free(out->write_work_buffer);
    free(out);

    ALOGD("%s() end", __FUNCTION__);
}

void close_stream_icx(stream_out* out)
{
    ALOGD("%s() start", __FUNCTION__);

#ifdef ENABLE_DATA_DUMP
    close_dump_file(DTYPE_ICX);
#endif

    finalize_stream_out(out);
    finalize_SRC(out); /* SRC only used in icx path. */
    free(out);

    ALOGD("%s() end", __FUNCTION__);
}

/*----------------------------------------------------------------------------*/
struct audio_stream_out*
sout_get_audio_stream_out(stream_out* out)
{
    HAL_ASSERT(out != NULL);
    return &(out->super);
}

bool
sout_is_standby(stream_out* out)
{
    if (out == NULL) { return true; }

    bool is_standby;

    pthread_mutex_lock(&out->stream_lock);
    is_standby = out->is_standby;
    pthread_mutex_unlock(&out->stream_lock);

    return is_standby;
}

void
sout_set_standby(stream_out* out,
                 bool        value)
{
    if (out == NULL) { return; }

    pthread_mutex_lock(&out->stream_lock);
    out->is_standby = value;
    ALOGD("standby becomes %s [%d]", (out->is_standby)?"true":"false", out->stype);
    pthread_mutex_unlock(&out->stream_lock);
}

stream_state
sout_get_state(stream_out* out)
{
    if (out == NULL) { return SS_NOT_USED; }

    bool     is_standby = sout_is_standby(out);
    uint32_t frame_num  = ICXB_get_readable_frame_num(out->icxbuf, false);

    if (is_standby)
    {
        if (frame_num == 0)
        {
            return SS_NOT_USED;
        }
        else
        {
            return SS_WAIT_CLOSING;
        }
    }
    else
    {
        if (frame_num == 0)
        {
            return SS_WAIT_PLAYING;
        }
        else
        {
            return SS_PLAYING;
        }
    }
}

void
sout_set_fade(stream_out* out,
              uint32_t    mute_time_ms)
{
    if (out == NULL) { return; }

    ALOGV("%s() mute_time_ms: %d", __FUNCTION__, mute_time_ms);
    set_fade_MUTE(out, mute_time_ms);
}

void
sout_flush_mute(stream_out* out)
{
    if (out == NULL) { return; }

    ALOGV("%s() [%d]", __FUNCTION__, out->stype);
    flush_MUTE(out);
}
