/*
 *  Copyright 2013 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU Lesser General Public License as published by
 *  the Free Software Foundation;  version 2.1 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 *
 */

/* #define TRACE_PRINT_ON */
/* #define DEBUG_PRINT_ON */
#define TRACE_TAG "        -------- "
#define DEBUG_TAG "                     "

#include "./common.h"

#include "misc.h"
#include "mute.h"

#define calc_vol(_v,_s,_n) ( ((_v)>>8) * (((_n)>>1) * 256 / ((_s)>>1)) )

static int elapse_time(uint64_t base);

struct mute_handle * mute_open(
	char * name,
	int    rate,
	int    fade
)
{
	struct mute_handle * handle;

	print_trace("%s(%s,%d,%d)",__FUNCTION__,name?name:"NULL",rate,fade);

	if(name==NULL){
		print_error("name is null.");
		return(NULL);
	}

	handle=malloc(sizeof(struct mute_handle));
	if(handle==NULL){
		print_fail("malloc(): no memory.");
		back_trace();
		return(NULL);
	}

	handle->name   = name;
	handle->rate   = rate;
	handle->fade   = fade;

	handle->virgin = TRUE;
	handle->base   = 0;

	handle->dir    = 1;
	handle->step   = handle->fade*handle->rate/1000+2;
	handle->now    = handle->step;

	handle->timeout = 0;
	handle->remain  = 0;

	pthread_mutex_init(&handle->mutex,NULL);

	return(handle);
}

void mute_close(
	struct mute_handle * handle
)
{
	print_trace("%s(%s)",__FUNCTION__,handle?handle->name:"NULL");

	if(handle!=NULL){
		pthread_mutex_destroy(&handle->mutex);
		free(handle);
	}

	return;
}

int mute_switch(
	struct mute_handle * handle,
	int                  value
)
{
	print_trace("%s(%s,%d)",__FUNCTION__,handle?handle->name:"NULL",value);

	pthread_mutex_lock(&handle->mutex);

	if(value)
		handle->dir=-1;
	else
		handle->dir=1;

	handle->base=system_time();

	pthread_mutex_unlock(&handle->mutex);

	return(0);
}

int mute_set_rate(
	struct mute_handle * handle,
	int                  rate
)
{
	print_trace("%s(%s,%d)",__FUNCTION__,handle?handle->name:"NULL",rate);

	pthread_mutex_lock(&handle->mutex);

	if(!handle->virgin){
		print_error("can not set rate.");
		pthread_mutex_unlock(&handle->mutex);
		return(-1);
	}

	handle->rate = rate;
	handle->step = handle->fade*handle->rate/1000+2;

	if(handle->now>0)
		handle->now = handle->step;

	pthread_mutex_unlock(&handle->mutex);

	return(0);
}

int mute_attenuate(
	struct mute_handle * handle,
	int *                buffer,
	int                  samples
)
{
	int offset;
	int elapse;

	print_trace("%s(%s,%d)",__FUNCTION__,handle?handle->name:"NULL",samples);

	pthread_mutex_lock(&handle->mutex);

	/* already unmute */
	/* if(handle->dir>0 && handle->now==handle->step){ */
		/* pthread_mutex_unlock(&handle->mutex); */
		/* return(0); */
	/* } */

	if(handle->virgin){
		print_debug("dir=%d, now=%d",handle->dir,handle->now);
	}

	/* fade in */
	if(handle->dir>0){

		if(handle->virgin && handle->now==0){
			elapse=elapse_time(handle->base);
			print_debug("elapse time = %d",elapse);
			elapse=minimum(elapse,handle->fade);
			elapse=maximum(elapse,0);
			handle->now=handle->step*elapse/handle->fade;
			print_debug("adjust %d -> %d",0,handle->now);
		}

		offset=0;

		if(handle->now<handle->step && samples>0){
			print_debug("fade-in");
		}

		while(handle->now<handle->step && samples>0){

			*(buffer+offset)=calc_vol(
				*(buffer+offset),
				handle->step,
				handle->now
			);

			samples--;
			offset++;

			handle->now++;
		}
	}

	/* fade out */
	else{

		if(handle->virgin && handle->now==handle->step){
			elapse=elapse_time(handle->base);
			print_debug("elapse time = %d",elapse);
			elapse=minimum(elapse,handle->fade);
			elapse=maximum(elapse,0);
			handle->now=handle->step-handle->step*elapse/handle->fade;
			print_debug("adjust %d -> %d",handle->step,handle->now);
		}

		offset=0;

		if(handle->now>0 && samples>0){
			print_debug("fade-out");
		}

		while(handle->now>0 && samples>0){

			handle->now--;

			*(buffer+offset)=calc_vol(
				*(buffer+offset),
				handle->step,
				handle->now
			);

			samples--;
			offset++;
		}

		memset(
			buffer+offset,
			0,
			samples*BYTE_PER_SAMPLE
		);
	}

	handle->virgin=FALSE;

	pthread_mutex_unlock(&handle->mutex);

	return(0);
}

int mute_flush(
	struct mute_handle * handle
)
{
	print_trace("%s(%s)",__FUNCTION__,handle?handle->name:"NULL");

	pthread_mutex_lock(&handle->mutex);

	if(handle->dir<0)
		handle->now = 0;
	else
		handle->now = handle->step;

	handle->virgin = TRUE;

	print_debug("dir=%d, now=%d",handle->dir,handle->now);

	pthread_mutex_unlock(&handle->mutex);

	return(0);
}

static int elapse_time(uint64_t base)
{
	uint64_t now;

	now=system_time();

	/* if(now<base) */
		print_error("n:%llu,b:%llu",now,base);

	return((int)((now-base)/1000000LL));
}









