/*
 *  Copyright 2013 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU Lesser General Public License as published by
 *  the Free Software Foundation;  version 2.1 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 *
 */

/* #define TRACE_PRINT_ON */
/* #define DEBUG_PRINT_ON */
#define TRACE_TAG "        -------- "
#define DEBUG_TAG "                     "

#include "./common.h"

#include "misc.h"
#include "mute.h"
#include "timed_mute.h"

static int elapse_time(uint64_t base);

struct timed_mute_handle * timed_mute_open(
	char * name,
	int    rate,
	int    fade
)
{
	struct timed_mute_handle * handle;

	print_trace("%s(%s,%d,%d)",__FUNCTION__,name?name:"NULL",rate,fade);

	if(name==NULL){
		print_error("name is null.");
		return(NULL);
	}

	handle=malloc(sizeof(struct timed_mute_handle));
	if(handle==NULL){
		print_fail("malloc(): no memory.");
		back_trace();
		return(NULL);
	}

	pthread_mutex_init(&handle->mutex,NULL);

	handle->handle=mute_open(name,rate,fade);
	if(handle->handle==NULL){
		back_trace();
		pthread_mutex_destroy(&handle->mutex);
		free(handle);
		return(NULL);
	}

	handle->name    = name;

	handle->rate    = rate;
	handle->mute    = FALSE;
	handle->virgin  = TRUE;

	handle->base    = 0;
	handle->timeout = 0;

	handle->remain  = 0;

	return(handle);
}

void timed_mute_close(
	struct timed_mute_handle * handle
)
{
	print_trace("%s(%s)",__FUNCTION__,handle?handle->name:"NULL");

	if(handle!=NULL){
		mute_close(handle->handle);
		pthread_mutex_destroy(&handle->mutex);
		free(handle);
	}

	return;
}

int timed_mute_set_rate(
	struct timed_mute_handle * handle,
	int                        rate
)
{
	int rv;

	print_trace("%s(%s,%d)",__FUNCTION__,handle?handle->name:"NULL",rate);

	pthread_mutex_lock(&handle->mutex);

	if(!handle->virgin){
		print_error("can not set rate.");
		pthread_mutex_unlock(&handle->mutex);
		return(-1);
	}

	rv=mute_set_rate(handle->handle,rate);
	if(rv<0){
		back_trace();
		pthread_mutex_unlock(&handle->mutex);
		return(-1);
	}

	handle->rate = rate;

	pthread_mutex_unlock(&handle->mutex);

	return(0);
}

int timed_mute_switch(
	struct timed_mute_handle * handle,
	int                        value
)
{
	uint64_t b;
	uint64_t t;
	int r;

	print_trace("%s(%s,%d)",__FUNCTION__,handle?handle->name:"NULL",value);

	pthread_mutex_lock(&handle->mutex);

	/* mute */
	if(value > 0){

		/* switch */
		if(handle->mute==FALSE){
			mute_switch(handle->handle,TRUE);
			handle->mute    = TRUE;
			handle->base    = system_time();
			handle->timeout = (uint64_t)value * 1000000LL;
			if(handle->virgin)
				handle->remain  = 0;
			else
				handle->remain  = (handle->rate * value / 1000) * 2;
			print_debug(
				"on: virgin=%d, base=%llu, tout=%llu, remain=%d",
				handle->virgin,
				handle->base,
				handle->timeout,
				handle->remain
			);
		}

		/* overwrite */
		else{

			if(handle->virgin){
				b = system_time();
				t = (uint64_t)value*1000000LL;
				r = 0;
				if(t+b > handle->timeout+handle->base){
					handle->base    = b;
					handle->timeout = t;
					handle->remain  = r;
					print_debug(
						"ovwr: virgin=%d, base=%llu, tout=%llu, remain=%d",
						handle->virgin,
						handle->base,
						handle->timeout,
						handle->remain
					);
				}
			}

			else{
				b = system_time();
				t = (uint64_t)value * 1000000LL;
				r = (handle->rate * value / 1000) * 2;
				if(r > handle->remain){
					handle->base    = b;
					handle->timeout = t;
					handle->remain  = r;
					print_debug(
						"ovwr: virgin=%d, base=%llu, tout=%llu, remain=%d",
						handle->virgin,
						handle->base,
						handle->timeout,
						handle->remain
					);
				}
			}
		}
	}

	/* unmute */
	else{
		if(handle->mute==TRUE){
			mute_switch(handle->handle,FALSE);
			handle->mute    = FALSE;
			handle->base    = 0;
			handle->timeout = 0;
			handle->remain  = 0;
			print_debug(
				"off: virgin=%d",
				handle->virgin
			);
		}
	}

	pthread_mutex_unlock(&handle->mutex);

	return(0);
}

int timed_mute_attenuate(
	struct timed_mute_handle * handle,
	int *                      buffer,
	int                        samples
)
{
	uint64_t now;
	int delta;
	int rv;

	print_trace("%s(%s,%d)",__FUNCTION__,handle?handle->name:"NULL",samples);

	pthread_mutex_lock(&handle->mutex);

	if(handle->mute==TRUE){

		if(handle->virgin){

			now = system_time();

			/* expire */
			if(now - handle->base >= handle->timeout){
				print_debug(
					"expire, now=%llu, base=%llu, tout=%llu, remain=%d",
					now,
					handle->base,
					handle->timeout,
					handle->remain
				);
				mute_switch(handle->handle,FALSE);
				mute_flush(handle->handle);
				handle->mute    = FALSE;
				handle->base    = 0;
				handle->timeout = 0;
				handle->remain  = 0;
			}

			/* adjust */
			else{
				delta = (handle->timeout + handle->base - now) / 1000000LL;
				handle->remain = (handle->rate * delta / 1000) * 2;
				print_debug(
					"adjust, now=%llu, base=%llu, tout=%llu, remain=%d",
					now,
					handle->base,
					handle->timeout,
					handle->remain
				);
			}
		}
	}

	rv=mute_attenuate(handle->handle,buffer,samples);
	if(rv<0){
		back_trace();
		pthread_mutex_unlock(&handle->mutex);
		return(-1);
	}

	if(handle->mute==TRUE){

		handle->remain = handle->remain - samples;

		if(handle->remain <= 0){
			print_debug(
				"expire, now=%llu, base=%llu, tout=%llu, remain=%d",
				now,
				handle->base,
				handle->timeout,
				handle->remain
			);
			mute_switch(handle->handle,FALSE);
			handle->mute    = FALSE;
			handle->base    = 0;
			handle->timeout = 0;
			handle->remain  = 0;
		}
	}

	handle->virgin = FALSE;

	pthread_mutex_unlock(&handle->mutex);

	return(0);
}

int timed_mute_flush(
	struct timed_mute_handle * handle
)
{
	int rv;

	print_trace("%s(%s)",__FUNCTION__,handle?handle->name:"NULL");

	pthread_mutex_lock(&handle->mutex);

	rv=mute_flush(handle->handle);
	if(rv<0){
		back_trace();
		pthread_mutex_unlock(&handle->mutex);
		return(-1);
	}

	handle->virgin = TRUE;

	pthread_mutex_unlock(&handle->mutex);

	return(0);
}


